/********************************************************
Function: lchol3(int n, int nb, double *A)
          performs the cholesky factorization for a lower
          triangular input matrix using matrix-matrix 
          level 3 blas routines in GSL and Fortran

Input arguments:

n:      the leading dimension of the array

nb:     the block size to be used 

A:      two-dimensional input array

On exit:  
        returns the modified matrix A

written by: kaska
date: nov. 23, 2004
********************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<gsl/gsl_vector.h>
#include<gsl/gsl_matrix.h>
#include<gsl/gsl_blas.h>
#include "lchol3.h"
#include "lchol2.h"

//definition of Fortran routines
void dtrsm_(char *side, char *uplo, char *transa, char *diag, int *m, int *n, double *alpha, double *a, int *lda, double *b, int *ldb);

void dsyrkkaska_(char *uplo, char *trans, int *N, int *K, double *alpha, double *A, int *lda, double *beta, double *C, int *ldc, int *tdah, int *tdch);


int lchol3(int n, int nb, double *A)
{
  int i,j;
  int b, bs;
  int one=1;
  int tdah, tdch;
  double alpha=-1.0;
  double beta=1.0;
  gsl_matrix *A11, *A21, *A22;
  gsl_matrix_view Agsl, A11gsl, A21gsl, A22gsl;
  char isRunning = 1;

  //using matrix views to avoid copying matrices
  Agsl = gsl_matrix_view_array(A, n, n);

  for (i = 0; i < n; i += nb) {

    b = n-i;

    if (i != 0) {
      A22gsl = gsl_matrix_submatrix(&Agsl.matrix, i, i, n-i, n-i);
      A21 = &A21gsl.matrix;
      A22 = &A22gsl.matrix;

      tdah = A21->tda - A21->size2;
      tdch = A22->tda - A22->size2;

      //could call dgemm instead of dsyrk
      //gsl_blas_dgemm(CblasNoTrans, CblasTrans, -1.0, &A21gsl.matrix, &A21gsl.matrix, 1.0, &A22gsl.matrix);
      //gsl_blas_dsyrk has a bug, use this call when new version becomes available
      //gsl_blas_dsyrk(CblasLower, CblasNoTrans, -1.0, &A21gsl.matrix, 1.0, &A22gsl.matrix);
      //calls the modified fortran blas 3 dsyrk function
      dsyrkkaska_("U", "T", &b, &nb, &alpha, A21->data, &nb, &beta, A22->data, &b, &tdah, &tdch);
    }
 
    bs = (nb < b ? nb : b);
    b -= nb;

    //level 2 cholesky call 
    A11gsl = gsl_matrix_submatrix(&Agsl.matrix, i, i, bs, bs);
    lchol2(bs,bs, &A11gsl.matrix);

    //dtrsm call
    if (b > 0 ) {
      A21gsl = gsl_matrix_submatrix(&Agsl.matrix, i+bs, i, n-i-bs, bs);
      gsl_blas_dtrsm(CblasRight, CblasLower, CblasTrans, CblasNonUnit, 1.0, &A11gsl.matrix, &A21gsl.matrix);
    }

  }

  return !isRunning;
}

